"""Provides security-related functions for the API."""

from .configs import JWTBearerAuthenticationConfig
from .decoders import JWTBearerTokenDecoder, JWTBearerTokenDecoderAbstract, decode_jwt_token_payload
from .exceptions import (
    InvalidJWTError,
    InvalidJWTPayploadError,
    JWTAuthenticationError,
    MissingJWTCredentialsError,
    NotVerifiedJWTError,
)
from .objects import JWTPayload
from .services import (
    JWTAuthenticationService,
    JWTAuthenticationServiceAbstract,
)
from .stores import JWKStoreAbstract, JWKStoreMemory
from .types import JWTToken, OAuth2Audience, OAuth2Issuer, OAuth2Scope, OAuth2Subject
from .verifiers import JWTNoneVerifier, JWTVerifierAbstract

__all__: list[str] = [
    "InvalidJWTError",
    "InvalidJWTPayploadError",
    "JWKStoreAbstract",
    "JWKStoreMemory",
    "JWTAuthenticationError",
    "JWTAuthenticationService",
    "JWTAuthenticationServiceAbstract",
    "JWTBearerAuthenticationConfig",
    "JWTBearerTokenDecoder",
    "JWTBearerTokenDecoderAbstract",
    "JWTNoneVerifier",
    "JWTPayload",
    "JWTToken",
    "JWTVerifierAbstract",
    "MissingJWTCredentialsError",
    "NotVerifiedJWTError",
    "OAuth2Audience",
    "OAuth2Issuer",
    "OAuth2Scope",
    "OAuth2Subject",
    "decode_jwt_token_payload",
]



"""Provides the JWT bearer authentication service."""

from http import HTTPStatus
from typing import Generic, TypeVar

from fastapi import HTTPException, Request

from fastapi_factory_utilities.core.security.abstracts import AuthenticationAbstract

from .configs import JWTBearerAuthenticationConfig
from .decoders import JWTBearerTokenDecoder, JWTBearerTokenDecoderAbstract
from .exceptions import InvalidJWTError, InvalidJWTPayploadError, MissingJWTCredentialsError, NotVerifiedJWTError
from .objects import JWTPayload
from .stores import JWKStoreAbstract
from .types import JWTToken
from .verifiers import JWTNoneVerifier, JWTVerifierAbstract

JWTBearerPayloadGeneric = TypeVar("JWTBearerPayloadGeneric", bound=JWTPayload)


class JWTAuthenticationServiceAbstract(AuthenticationAbstract, Generic[JWTBearerPayloadGeneric]):
    """JWT authentication service.

    This service is the orchestrator for the JWT bearer authentication.
    """

    def __init__(
        self,
        jwt_bearer_authentication_config: JWTBearerAuthenticationConfig,
        jwt_verifier: JWTVerifierAbstract[JWTBearerPayloadGeneric],
        jwt_decoder: JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric],
        raise_exception: bool = True,
    ) -> None:
        """Initialize the JWT bearer authentication service.

        Args:
            jwt_bearer_authentication_config (JWTBearerAuthenticationConfig): The JWT bearer authentication
            configuration.
            jwt_verifier (JWTVerifierAbstract): The JWT bearer token verifier.
            jwt_decoder (JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]): The JWT bearer token decoder.
            raise_exception (bool, optional): Whether to raise an exception or return None. Defaults to True.
        """
        # Configuration and Behavior
        self._jwt_bearer_authentication_config: JWTBearerAuthenticationConfig = jwt_bearer_authentication_config
        self._jwt_verifier: JWTVerifierAbstract[JWTBearerPayloadGeneric] = jwt_verifier
        self._jwt_decoder: JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric] = jwt_decoder
        # Runtime variables
        self._jwt: JWTToken | None = None
        self._jwt_payload: JWTBearerPayloadGeneric | None = None
        super().__init__(raise_exception=raise_exception)

    @property
    def verifier(self) -> JWTVerifierAbstract[JWTBearerPayloadGeneric]:
        """Get the JWT bearer token verifier.

        Returns:
            JWTVerifierAbstract[JWTBearerPayloadGeneric]: The JWT bearer token verifier.
        """
        return self._jwt_verifier

    @property
    def decoder(self) -> JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]:
        """Get the JWT bearer token decoder.

        Returns:
            JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]: The JWT bearer token decoder.
        """
        return self._jwt_decoder

    @classmethod
    def extract_authorization_header_from_request(cls, request: Request) -> str:
        """Extract the authorization header from the request.

        Args:
            request (Request): The request object.

        Returns:
            str: The authorization header.

        Raises:
            MissingJWTCredentialsError: If the authorization header is missing.
        """
        authorization_header: str | None = request.headers.get("Authorization", None)
        if not authorization_header:
            raise MissingJWTCredentialsError(message="Missing Credentials")
        return authorization_header

    @classmethod
    def extract_bearer_token_from_authorization_header(cls, authorization_header: str) -> JWTToken:
        """Extract the bearer token from the authorization header.

        Args:
            authorization_header (str): The authorization header.

        Returns:
            JWTToken: The bearer token.

        Raises:
            InvalidJWTError: If the authorization header is invalid.
        """
        if not authorization_header.startswith("Bearer "):
            raise InvalidJWTError(message="Invalid Credentials")
        return JWTToken(authorization_header.split(sep=" ")[1])

    @property
    def payload(self) -> JWTBearerPayloadGeneric | None:
        """Get the JWT bearer payload.

        Returns:
            JWTBearerPayloadGeneric | None: The JWT bearer payload, or None if not authenticated yet.
        """
        return self._jwt_payload

    async def authenticate(self, request: Request) -> None:
        """Authenticate the JWT bearer token.

        Args:
            request (Request): The request object.

        Returns:
            None: If the authentication is successful or not raise_exception is False.

        Raises:
            MissingJWTCredentialsError: If the authorization header is missing.
            InvalidJWTError: If the authorization header is invalid.
            InvalidJWTPayploadError: If the JWT bearer token payload is invalid.
            NotVerifiedJWTError: If the JWT bearer token is not verified.
        """
        authorization_header: str
        try:
            authorization_header = self.extract_authorization_header_from_request(request=request)
            self._jwt = self.extract_bearer_token_from_authorization_header(authorization_header=authorization_header)
        except (MissingJWTCredentialsError, InvalidJWTError) as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail=str(e)))

        try:
            self._jwt_payload = await self._jwt_decoder.decode_payload(jwt_token=self._jwt)
        except (InvalidJWTError, InvalidJWTPayploadError) as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.FORBIDDEN, detail=str(e)))

        try:
            await self._jwt_verifier.verify(jwt_token=self._jwt, jwt_payload=self._jwt_payload)
        except NotVerifiedJWTError as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.FORBIDDEN, detail=str(e)))

        return


class JWTAuthenticationService(JWTAuthenticationServiceAbstract[JWTPayload]):
    """JWT bearer authentication service."""

    def __init__(
        self,
        jwt_bearer_authentication_config: JWTBearerAuthenticationConfig,
        jwks_store: JWKStoreAbstract,
        raise_exception: bool = True,
    ) -> None:
        """Initialize the JWT bearer authentication service.

        Don't enforce the public_key from configuration, for the developper to
        provide it through the dependency injection freely from any source.

        Args:
            jwt_bearer_authentication_config (JWTBearerAuthenticationConfig): The JWT bearer authentication
            configuration.
            jwks_store (JWKStoreAbstract): The JWKS store.
            raise_exception (bool, optional): Whether to raise an exception or return None. Defaults to True.
        """
        super().__init__(
            jwt_bearer_authentication_config=jwt_bearer_authentication_config,
            jwt_verifier=JWTNoneVerifier(),
            jwt_decoder=JWTBearerTokenDecoder(
                jwt_bearer_authentication_config=jwt_bearer_authentication_config, jwks_store=jwks_store
            ),
            raise_exception=raise_exception,
        )



"""Provides the configurations for the JWT bearer token."""

from typing import ClassVar

from jwt.algorithms import get_default_algorithms, requires_cryptography
from pydantic import BaseModel, ConfigDict, Field, field_validator


class JWTBearerAuthenticationConfig(BaseModel):
    """JWT bearer token authentication configuration."""

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    authorized_algorithms: list[str] = Field(
        default_factory=lambda: list(get_default_algorithms().keys()), description="The authorized algorithms."
    )

    audience: str = Field(description="The audience to be included in the JWT token.")
    authorized_audiences: list[str] | None = Field(default=None, description="The authorized audiences.")
    authorized_issuers: list[str] | None = Field(default=None, description="The authorized issuers.")

    @field_validator("authorized_algorithms")
    @classmethod
    def validate_authorized_algorithms(cls, v: list[str]) -> list[str]:
        """Validate the authorized algorithms."""
        invalid_algorithms: list[str] = []
        for algorithm in v:
            if algorithm not in requires_cryptography:
                invalid_algorithms.append(algorithm)
        if invalid_algorithms:
            raise ValueError(f"Invalid algorithms: {invalid_algorithms}")
        return v



"""FastAPI Factory Utilities exceptions."""

import logging
import traceback
from typing import Any, cast

from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv.attributes.exception_attributes import (
    EXCEPTION_MESSAGE,
    EXCEPTION_STACKTRACE,
    EXCEPTION_TYPE,
)
from opentelemetry.trace import Span, get_current_span
from opentelemetry.util.types import AttributeValue
from structlog.stdlib import BoundLogger, get_logger

_logger: BoundLogger = get_logger()


class FastAPIFactoryUtilitiesError(Exception):
    """Base exception for the FastAPI Factory Utilities."""

    FILTERED_ATTRIBUTES: tuple[str, ...] = ()
    DEFAULT_LOGGING_LEVEL: int = logging.ERROR
    DEFAULT_MESSAGE: str | None = None

    def __init__(
        self,
        *args: object,
        **kwargs: Any,
    ) -> None:
        """Instantiate the exception.

        Args:
            *args: The arguments.
            **kwargs: The keyword arguments.

        """
        # If Default Message is not set, try to extract it from docstring (first line)
        default_message: str = self.DEFAULT_MESSAGE or "An error occurred"
        if self.DEFAULT_MESSAGE is None and self.__doc__ is not None:
            default_message = self.__doc__.split("\n", maxsplit=1)[0]
        # Extract the message and the level from the kwargs if they are present
        self.message: str | None = cast(str | None, kwargs.pop("message", None))
        self.level: int = cast(int, kwargs.pop("level", self.DEFAULT_LOGGING_LEVEL))

        # If the message is not present, try to extract it from the args
        if self.message is None and len(args) > 0 and isinstance(args[0], str):
            self.message = args[0]
        elif self.message is None:
            self.message = default_message

        # Log the Exception
        if self.message:
            _logger.log(level=self.level, event=self.message)

        # Set the kwargs as attributes of the exception
        for key, value in kwargs.items():
            if key in self.FILTERED_ATTRIBUTES:
                continue
            setattr(self, key, value)

        try:
            # Propagate the exception
            span: Span = get_current_span()
            # If not otel is setup, INVALID_SPAN is retrieved from get_current_span
            # and it will respond False to the is_recording method
            if span.is_recording():
                # Set the kwargs attributes
                for key, value in kwargs.items():
                    if key in self.FILTERED_ATTRIBUTES:
                        continue
                    attribute_value: AttributeValue
                    if not isinstance(value, (str, bool, int, float)):
                        attribute_value = str(value)
                    else:
                        attribute_value = value
                    span.set_attribute(key, attribute_value)

                # Record official Attributes last to avoid overriding them
                span.record_exception(self)
                # Set the exception and error attributes
                span.set_attribute(ERROR_TYPE, self.__class__.__name__)
                span.set_attribute(EXCEPTION_MESSAGE, self.message)
                span.set_attribute(EXCEPTION_STACKTRACE, traceback.format_exc())
                span.set_attribute(EXCEPTION_TYPE, self.__class__.__name__)
        except Exception:  # pylint: disable=broad-exception-caught
            # Suppress any errors that occur while propagating the exception
            pass

        # Call the parent class
        super().__init__(*args)

    def __str__(self) -> str:
        """Return the string representation of the exception.

        Returns:
            str: The message if available, otherwise the default exception string.
        """
        if self.message is not None:
            return self.message
        return super().__str__()



"""Provides a service to interact with the Hydra service."""

import json
from base64 import b64encode
from http import HTTPStatus
from typing import Annotated, Any, Generic, TypeVar, get_args

import aiohttp
import jwt
from fastapi import Depends
from pydantic import ValidationError

from fastapi_factory_utilities.core.plugins.aiohttp import AioHttpClientResource, AioHttpResourceDepends

from .exceptions import HydraOperationError
from .objects import HydraTokenIntrospectObject
from .types import HydraAccessToken, HydraClientId, HydraClientSecret

HydraIntrospectObjectGeneric = TypeVar("HydraIntrospectObjectGeneric", bound=HydraTokenIntrospectObject)


class HydraIntrospectGenericService(Generic[HydraIntrospectObjectGeneric]):
    """Service to interact with the Hydra introspect service with a generic introspect object."""

    INTROSPECT_ENDPOINT: str = "/admin/oauth2/introspect"
    WELLKNOWN_JWKS_ENDPOINT: str = "/.well-known/jwks.json"

    def __init__(
        self,
        hydra_admin_http_resource: AioHttpClientResource,
        hydra_public_http_resource: AioHttpClientResource,
    ) -> None:
        """Instanciate the Hydra introspect service.

        Args:
            hydra_admin_http_resource (AioHttpClientResource): The Hydra admin HTTP resource.
            hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.
        """
        self._hydra_admin_http_resource: AioHttpClientResource = hydra_admin_http_resource
        self._hydra_public_http_resource: AioHttpClientResource = hydra_public_http_resource
        # Retrieve the concrete introspect object class
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._concreate_introspect_object_class: type[HydraIntrospectObjectGeneric] = generic_args[0]

    async def introspect(self, token: HydraAccessToken) -> HydraIntrospectObjectGeneric:
        """Introspects a token using the Hydra introspect service.

        Args:
            token (str): The token to introspect.
        """
        try:
            async with self._hydra_admin_http_resource.acquire_client_session() as session:
                async with session.post(
                    url=self.INTROSPECT_ENDPOINT,
                    data={"token": token},
                ) as response:
                    response.raise_for_status()
                    instrospect: HydraIntrospectObjectGeneric = self._concreate_introspect_object_class.model_validate(
                        await response.json()
                    )
        except aiohttp.ClientResponseError as error:
            raise HydraOperationError("Failed to introspect the token", status_code=error.status) from error
        except json.JSONDecodeError as error:
            raise HydraOperationError("Failed to decode the introspect response") from error
        except ValidationError as error:
            raise HydraOperationError("Failed to validate the introspect response") from error

        return instrospect

    async def get_wellknown_jwks(self) -> jwt.PyJWKSet:
        """Get the JWKS from the Hydra service."""
        try:
            async with self._hydra_public_http_resource.acquire_client_session() as session:
                async with session.get(
                    url=self.WELLKNOWN_JWKS_ENDPOINT,
                ) as response:
                    response.raise_for_status()
                    jwks_data: dict[str, Any] = await response.json()
                    jwks: jwt.PyJWKSet = jwt.PyJWKSet.from_dict(jwks_data)
                    return jwks
        except aiohttp.ClientResponseError as error:
            raise HydraOperationError(
                "Failed to get the JWKS from the Hydra service", status_code=error.status
            ) from error
        except json.JSONDecodeError as error:
            raise HydraOperationError("Failed to decode the JWKS from the Hydra service") from error
        except ValidationError as error:
            raise HydraOperationError("Failed to validate the JWKS from the Hydra service") from error


class HydraIntrospectService(HydraIntrospectGenericService[HydraTokenIntrospectObject]):
    """Service to interact with the Hydra introspect service with the default HydraTokenIntrospectObject."""


class HydraOAuth2ClientCredentialsService:
    """Service to interact with the Hydra service."""

    INTROSPECT_ENDPOINT: str = "/admin/oauth2/introspect"
    CLIENT_CREDENTIALS_ENDPOINT: str = "/oauth2/token"

    def __init__(
        self,
        hydra_public_http_resource: AioHttpClientResource,
    ) -> None:
        """Instanciate the Hydra service.

        Args:
            hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.
        """
        self._hydra_public_http_resource: AioHttpClientResource = hydra_public_http_resource

    @classmethod
    def build_bearer_header(cls, client_id: HydraClientId, client_secret: HydraClientSecret) -> str:
        """Build the bearer header.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.

        Returns:
            str: The bearer header.
        """
        auth_string: str = f"{client_id}:{client_secret}"
        auth_bytes: bytes = auth_string.encode("utf-8")
        auth_b64: str = b64encode(auth_bytes).decode("utf-8")
        return f"Basic {auth_b64}"

    async def oauth2_client_credentials(
        self, client_id: HydraClientId, client_secret: HydraClientSecret, scopes: list[str]
    ) -> HydraAccessToken:
        """Get the OAuth2 client credentials.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.
            scopes (list[str]): The scopes

        Returns:
            str: The access token.

        Raises:
            HydraOperationError: If the client credentials request fails.
        """
        async with self._hydra_public_http_resource.acquire_client_session() as session:
            async with session.post(
                url=self.CLIENT_CREDENTIALS_ENDPOINT,
                headers={"Authorization": self.build_bearer_header(client_id, client_secret)},
                data={"grant_type": "client_credentials", "scope": " ".join(scopes)},
            ) as response:
                response_data = await response.json()
                if response.status != HTTPStatus.OK:
                    raise HydraOperationError(f"Failed to get client credentials: {response_data}")

                return response_data["access_token"]


def depends_hydra_oauth2_client_credentials_service(
    hydra_public_http_resource: Annotated[AioHttpClientResource, Depends(AioHttpResourceDepends("hydra_public"))],
) -> HydraOAuth2ClientCredentialsService:
    """Dependency injection for the Hydra OAuth2 client credentials service.

    Args:
        hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.

    Returns:
        HydraOAuth2ClientCredentialsService: The Hydra OAuth2 client credentials service instance.

    Raises:
        HydraOperationError: If the Hydra public dependency is not configured.
    """
    return HydraOAuth2ClientCredentialsService(
        hydra_public_http_resource=hydra_public_http_resource,
    )


def depends_hydra_introspect_service(
    hydra_admin_http_resource: Annotated[AioHttpClientResource, Depends(AioHttpResourceDepends("hydra_admin"))],
    hydra_public_http_resource: Annotated[AioHttpClientResource, Depends(AioHttpResourceDepends("hydra_public"))],
) -> HydraIntrospectService:
    """Dependency injection for the Hydra introspect service.

    Args:
        hydra_admin_http_resource (AioHttpClientResource): The Hydra admin HTTP resource.
        hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.

    Returns:
        HydraIntrospectService: The Hydra introspect service instance.

    Raises:
        HydraOperationError: If the Hydra admin dependency is not configured.
    """
    return HydraIntrospectService(
        hydra_admin_http_resource=hydra_admin_http_resource,
        hydra_public_http_resource=hydra_public_http_resource,
    )



"""Provides the JWK stores."""

from abc import ABC, abstractmethod
from asyncio import Lock

from jwt.api_jwk import PyJWK, PyJWKSet


class JWKStoreAbstract(ABC):
    """JWK store abstract class."""

    async def get_jwk(self, kid: str) -> PyJWK:
        """Get the JWK from the store."""
        return (await self.get_jwks())[kid]

    @abstractmethod
    async def get_jwks(self) -> PyJWKSet:
        """Get the JWKS from the store."""
        raise NotImplementedError()

    @abstractmethod
    async def store_jwks(self, jwks: PyJWKSet) -> None:
        """Store the JWKS in the store."""
        raise NotImplementedError()


class JWKStoreMemory(JWKStoreAbstract):
    """JWK store in memory. Concurrent safe."""

    def __init__(self) -> None:
        """Initialize the JWK store in memory."""
        self._jwks: PyJWKSet
        self._lock: Lock = Lock()

    async def get_jwks(self) -> PyJWKSet:
        """Get the JWKS from the store."""
        async with self._lock:
            return self._jwks

    async def store_jwks(self, jwks: PyJWKSet) -> None:
        """Store the JWKS in the store."""
        async with self._lock:
            self._jwks = jwks
